'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var util = require('util');
var glob = _interopDefault(require('glob'));
var path = _interopDefault(require('path'));
var fs = _interopDefault(require('fs'));

var Exception = /*@__PURE__*/(function (Error) {
  function Exception(message, e) {
    Error.call(this, message);
    this.stack = e && e.stack || this.stack;
  }

  if ( Error ) Exception.__proto__ = Error;
  Exception.prototype = Object.create( Error && Error.prototype );
  Exception.prototype.constructor = Exception;

  return Exception;
}(Error));

var RESOLVED_STATE = Symbol('@@Resolved');
var IS_INJECTABLE = Symbol('@@Injector');

function info(value, prop) {
  return Object.getOwnPropertyDescriptor(value, prop);
}

function isClass(value) {
  var descriptor = info(value, 'prototype');
  return descriptor && descriptor.writable === false || value.toString().indexOf('class') === 0;
}

function isAsync(value) {
  return Object.prototype.toString.call(value) === '[object AsyncFunction]';
}

function classInjector(Factory) {
  return function (deps) { return new Factory(deps); };
}

function functionInjector(definition) {
  /* istanbul ignore if */
  if (isAsync(definition)) {
    return function () { return definition; };
  }

  return typeof definition.prototype !== 'undefined' ? function () { return definition; } : definition;
}

var Injector = function Injector(definition, injectables) {
  if (typeof definition !== 'function') {
    throw new Exception(("Cannot inject non-callable values, given " + (util.inspect(definition))));
  }

  if (!injectables || Array.isArray(injectables) || typeof injectables !== 'object') {
    throw new Exception(("Invalid injectables, given " + (util.inspect(injectables))));
  }

  var _isClass = isClass(definition);

  Object.defineProperty(this, '_class', {
    enumerable: false,
    value: _isClass
  });
  Object.defineProperty(this, '_factory', {
    enumerable: false,
    value: _isClass ? classInjector(definition) : functionInjector(definition)
  });
  Object.defineProperty(this, '_definition', {
    enumerable: false,
    value: definition
  });
  Object.defineProperty(this, '_dependencies', {
    enumerable: false,
    value: injectables
  });
};

var prototypeAccessors = { isClass: { configurable: true } };
var staticAccessors = { Symbol: { configurable: true } };

staticAccessors.Symbol.get = function () {
  return IS_INJECTABLE;
};

Injector.hasLocked = function hasLocked (target) {
  return !!target[RESOLVED_STATE];
};

Injector.supports = function supports (target) {
  return target instanceof Injector;
};

Injector.assign = function assign (target, extensions) {
  if (['object', 'function'].indexOf(typeof target) === -1 || Array.isArray(target)) {
    throw new Exception(("Target is not an object, given " + (util.inspect(target))));
  }

  if (target[RESOLVED_STATE]) {
    throw new Exception('Cannot assign to locked values');
  }

  Object.assign(target, extensions);
  Object.defineProperty(target, RESOLVED_STATE, {
    enumerable: false,
    value: true
  });
  return target;
};

Injector.bind = function bind (resolver, definition, defCallbacks) {
  var keys = Object.keys(resolver.values);
  var deps = definition._dependencies;
  var wrap = definition._factory;
  var resolved = {};
  var values = {};
  var proxy = {};
  Object.getOwnPropertyNames(deps).forEach(function (key) {
    var propName = key.replace(/^get/, '');

    if (keys.indexOf(propName) === -1) {
      throw new Exception(("Missing '" + propName + "' dependency"));
    }

    if (!resolver.values[propName]) {
      throw new Exception(("Value '" + propName + "' is not defined"));
    }

    var descriptor = info(deps, key);
    var isGetter = !!descriptor.get;
    var method = descriptor.get || descriptor.value;

    if (typeof method !== 'function') {
      throw new Exception(("Invalid resolver, given " + (util.inspect(method))));
    }

    if (isGetter) {
      values[propName] = method.bind(resolver.valueOf());
      return;
    }

    values[propName] = function () {
      if (typeof resolved[propName] === 'undefined') {
        var newValue = method.call(resolver.valueOf(), proxy);
        resolved[propName] = newValue || resolver.get(propName, defCallbacks);
      }

      return resolved[propName];
    };
  });
  keys.forEach(function (key) {
    Object.defineProperty(proxy, key, {
      enumerable: false,
      get: function () {
        if (!values[key]) {
          throw new Exception(("Missing '" + key + "' provider"));
        }

        return values[key]();
      }
    });
  });
  return wrap(proxy);
};

Injector.use = function use (resolver, definition) {
  var deps = definition._dependencies;
  var wrap = definition._factory;
  var values = {};
  var proxy = {};
  Object.getOwnPropertyNames(deps).forEach(function (key) {
    var propName = key.replace(/^get/, '');
    Object.defineProperty(proxy, propName, {
      get: function () {
        if (typeof values[key] === 'undefined') {
          if (typeof deps[key] !== 'function') {
            throw new Exception(("Invalid resolver, given " + (util.inspect(deps[key]))));
          }

          values[key] = deps[key].call(resolver.valueOf()) || null;
        }

        return values[key];
      }
    });
  });
  return wrap(proxy);
};

prototypeAccessors.isClass.get = function () {
  return this._class;
};

Injector.prototype.valueOf = function valueOf () {
  return this._definition;
};

Object.defineProperties( Injector.prototype, prototypeAccessors );
Object.defineProperties( Injector, staticAccessors );

var Container = function Container(rootContainer, dependencies) {
  Object.assign(this, dependencies);
  Object.defineProperty(this, '_root', {
    enumerable: false,
    value: rootContainer
  });
  Object.defineProperty(this, '_lock', {
    enumerable: false,
    value: {}
  });
};

Container.unwrap = function unwrap (resolver, definition, hooks) {
  if (!definition || typeof definition === 'function' || typeof definition !== 'object') {
    return definition;
  }

  var target = {};
  Object.keys(definition).forEach(function (propName) {
    var value = definition[propName];

    if (Injector.supports(value)) {
      target[propName] = Injector.bind(resolver, value, hooks);
    } else {
      target[propName] = Container.unwrap(resolver, value, hooks);
    }
  });
  return target;
};

Container.prototype.valueOf = function valueOf () {
  return this._root;
};

Container.prototype.get = function get (value, hooks) {
  var target = this.values[value];

  if (target === Injector.Symbol) {
    return null;
  }

  if (!target || Array.isArray(target) || ['object', 'function'].indexOf(typeof target) === -1) {
    throw new Exception(("Target '" + value + "' is not an object, given " + (util.inspect(target))));
  }

  if (Injector.supports(target)) {
    if (target.isClass) {
      var Class = target.valueOf();

      if (Class.prototype.constructor.length !== 1) {
        return hooks && hooks.after(value, Class) || Class;
      }
    }

    target = Injector.use(this, target);
  }

  if (!(this._lock[value] || Injector.hasLocked(target))) {
    this._lock[value] = true;

    try {
      var retval = Injector.assign(target, Container.unwrap(this, this.registry[value], hooks));

      if (hooks && hooks.after) {
        retval = hooks.after(value, retval);
      }

      if (typeof retval !== 'undefined') {
        this.values[value] = retval;
      }

      return this.values[value];
    } catch (e) {
      throw new Exception(("Definition of '" + value + "' failed. " + (e.message)), e);
    }
  }

  return target;
};

var ALLOWED_HOOKS = ['before', 'after'];

function getDecorators(hooks, container) {
  if (typeof hooks === 'function') {
    var after = hooks;
    hooks = {
      after: after
    };
  }

  hooks = hooks || {};
  return ALLOWED_HOOKS.reduce(function (prev, cur) {
    prev[cur] = function call() {
      return hooks[cur] && hooks[cur].apply(container, arguments);
    };

    return prev;
  }, {});
}

function assignProps(target) {
  var sources = Array.prototype.slice.call(arguments, 1);
  sources.forEach(function (source) {
    if (source) {
      Object.getOwnPropertyNames(source).forEach(function (dep) {
        Object.defineProperty(target, dep, Object.getOwnPropertyDescriptor(source, dep));
      });
    }
  });
  return target;
}

function camelCase(value) {
  return value.replace(/-([a-z])/g, function (_, $1) { return $1.toUpperCase(); });
}

function ucFirst(value) {
  return value[0].toUpperCase() + value.substr(1);
}

var Resolver = function Resolver(rootContainer, directory, hooks) {
  if (typeof rootContainer === 'string') {
    hooks = directory;
    directory = rootContainer;
    rootContainer = undefined;
  }

  Object.defineProperty(this, '_decorators', {
    enumerable: false,
    value: getDecorators(hooks, rootContainer)
  });
  Object.defineProperty(this, '_container', {
    enumerable: false,
    value: new Container(rootContainer, Resolver.scanFiles(directory, this._decorators.before))
  });
};

var prototypeAccessors$1 = { values: { configurable: true },registry: { configurable: true } };

Resolver.scanFiles = function scanFiles (cwd, cb) {
  var resolverInfo = {
    registry: {},
    values: {}
  };
  var entryFiles = glob.sync('**/index.js', {
    cwd: cwd,
    nosort: true
  }).sort(function (a, b) { return a.split('/').length - b.split('/').length; }).filter(function (f) { return f !== 'index.js'; });
  var rootProvider = path.join(cwd, 'provider.js');
  var rootDependencies = Resolver.useFile(rootProvider);

  function injectDefinition(target, providerFile) {
    return new Injector(target, assignProps({}, rootDependencies, Resolver.useFile(providerFile)));
  }

  entryFiles.forEach(function (entry) {
    var properties = entry.split('/');
    var value = ucFirst(camelCase(properties.shift()));
    var definitionFile = path.join(cwd, entry);
    var definition = Resolver.loadFile(definitionFile);
    properties.pop();
    var providerFile = path.join(cwd, value, properties.concat('provider.js').join('/'));
    var isInjectable = typeof definition === 'function';

    if (!resolverInfo.values[value]) {
      resolverInfo.values[value] = !properties.length ? isInjectable && injectDefinition(definition, providerFile) || definition : {};
    }

    if (!isInjectable && fs.existsSync(providerFile)) {
      throw new Exception(("Unexpected provider file, given " + providerFile));
    }

    resolverInfo.registry[value] = resolverInfo.registry[value] || {};
    var target = resolverInfo.registry[value];
    var propName;

    while (properties.length) {
      propName = camelCase(properties.shift());

      if (!target[propName]) {
        if (isInjectable && !properties.length) {
          target[propName] = injectDefinition(definition, providerFile);
        } else {
          target = target[propName] = {};
        }
      } else {
        target = target[propName];
      }
    }
  });
  Object.keys(resolverInfo.values).forEach(function (prop) {
    var definition = resolverInfo.values[prop];

    if (typeof cb === 'function' && !Injector.supports(definition)) {
      resolverInfo.values[prop] = cb(prop, definition) || definition;
    }
  });
  Object.getOwnPropertyNames(rootDependencies).forEach(function (key) {
    var propName = key.replace(/^get/, '');

    if (typeof resolverInfo.values[propName] === 'undefined') {
      resolverInfo.values[propName] = Injector.Symbol;
    }
  });
  return resolverInfo;
};

Resolver.loadFile = function loadFile (definition) {
  return require(definition);
};

Resolver.useFile = function useFile (providerFile) {
  var dependencies = {};

  if (fs.existsSync(providerFile)) {
    assignProps(dependencies, Resolver.loadFile(providerFile));
  }

  return dependencies;
};

prototypeAccessors$1.values.get = function () {
  return this._container.values;
};

prototypeAccessors$1.registry.get = function () {
  return this._container.registry;
};

Resolver.prototype.get = function get (name) {
  return this._container.get(name, this._decorators);
};

Object.defineProperties( Resolver.prototype, prototypeAccessors$1 );

var Chainable = function Chainable(context, middleware) {
  if (!middleware || Array.isArray(middleware) || typeof middleware !== 'object') {
    throw new Exception(("Invalid middleware, given " + (util.inspect(middleware))));
  }

  var chain = [];

  function factory() {
    var arguments$1 = arguments;

    if (!chain.length) {
      throw new Exception('Missing middleware to chain');
    }

    return chain.reduce(function (prev, cur) { return prev.then(function () { return cur.apply(context, arguments$1); }); }, Promise.resolve());
  }

  Object.keys(middleware).forEach(function (prop) {
    if (typeof middleware[prop] !== 'function') {
      throw new Exception(("Invalid function for '" + prop + "', given " + (util.inspect(middleware[prop]))));
    }

    Object.defineProperty(factory, prop, {
      enumerable: false,
      get: function () {
        chain.push(middleware[prop]);
        return factory;
      }
    });
  });
  return function call() {
    chain = [];
    var deferred = Promise.resolve();
    Array.prototype.slice.call(arguments).forEach(function (cb) {
      if (typeof cb !== 'function') {
        throw new Exception(("Invalid factory, given " + (util.inspect(cb))));
      }

      deferred = cb(factory) || deferred;
    });
    return deferred;
  };
};

exports.Chainable = Chainable;
exports.Container = Container;
exports.Injector = Injector;
exports.Resolver = Resolver;
